(**************************************************************************

   	This file is part of the AtomicDensityMatrix package for Mathematica
    Copyright (C) 2004-2019 Simon M. Rochester <simon@rochesterscientific.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

**************************************************************************)


BeginPackage["AtomicDensityMatrix`AtomicSystem`", {"AtomicDensityMatrix`"}];

Begin["`Private`"];

AtomicState[label_, {rules___Rule}] := AtomicState[label, rules];

StateLabel[state_AtomicState] := First@state;
StateLabel[sys_?AtomicSystemQ] := StateLabel /@ sys;

StateParameters[state_AtomicState] ^:= Rest[List @@ state];

StateParameterPattern[] :=
    S | L | J | NuclearSpin | F | M | Parity | Energy | NaturalWidth |
        GFactor | BranchingRatio[_] | LevelPosition | HyperfineA |
        HyperfineB | Polarizability[_] | TransitionPolarizability[_];

StateParameterQ[expr_] := MatchQ[expr, StateParameterPattern[]];

AtomicSystemPattern[] = {_AtomicState..};

AtomicSystemQ[expr_] := MatchQ[expr, AtomicSystemPattern[]];

DefinedQ[state_AtomicState, x : StateParameterPattern[]] := MemberQ[Rest@state, x -> _];

DefinedQ[sys : _AtomicState | {_AtomicState ..}, x : {StateParameterPattern[] ..} | StateParameterPattern[] ..] :=
    And @@ Flatten@Outer[DefinedQ, Flatten@{sys}, Flatten@{x}];

DeleteParameter[sys : {_AtomicState ..} | _AtomicState, x : {StateParameterPattern[] ..} | StateParameterPattern[]..] :=
    DeleteParameter[sys, x] =
        DeleteCases[sys, Alternatives @@ Thread[Flatten[{x}] -> _], {ArrayDepth[{sys}]}];

AddParameter[state_AtomicState, rules : {(StateParameterPattern[] -> _) ..} | (StateParameterPattern[] -> _)..] :=
    Join[DeleteParameter[state, Flatten[{rules}][[All, 1]]], AtomicState @@ Flatten[{rules}]];
AddParameter[sys_?AtomicSystemQ, rules : {(StateParameterPattern[] -> _) ..} | (StateParameterPattern[] -> _)..] :=
    Thread[Unevaluated@AddParameter[sys, rules], List, 1];

With[{f = #},
    f[state_AtomicState] := f[state] = f /. StateParameters[state] /. f -> f[StateLabel[state]];
    f[sys_?AtomicSystemQ] := f /@ sys;
]& /@ {S, L, NuclearSpin, M, Parity, Energy, NaturalWidth, GFactor, LevelPosition, HyperfineA, HyperfineB};

J[state_AtomicState] := J@state = J /. StateParameters@state /. J -> 0;
J[sys_?AtomicSystemQ] := J /@ sys;

F[state_AtomicState] := F@state = F /. StateParameters@state /. F -> J@state;
F[sys_?AtomicSystemQ] := F /@ sys;

BranchingRatio[state1_AtomicState][state2_AtomicState] := If[
    (*if reverse branching ratio is nonzero, or states have the same parity, then this b.r. is zero*)
    (DefinedQ[state1, BranchingRatio[StateLabel@state2]] && BranchingRatio[StateLabel@state2][state1] =!= 0)
        || Parity@state1 === Parity@state2,
    0,
    (*else calculate b.r. using state label*)
    BranchingRatio[StateLabel@state1][state2]
];

BranchingRatio[label_][state_AtomicState] := Module[{value},
    If[
        StateLabel@state === label || NaturalWidth@state === 0,
        0,
        If[
            (value = BranchingRatio@label /. StateParameters@state) =!= BranchingRatio@label,
            value,
            BranchingRatio[StateLabel@state, label]
        ]
    ]
];

Polarizability[k__][state_AtomicState] := Module[{value},
    If[(value = Polarizability[k] /. StateParameters@state) =!= Polarizability[k],
        value,
        Polarizability[k][StateLabel@state]
    ]
];

TransitionPolarizability[k_][state1_AtomicState, state2_AtomicState] := Module[{value},
    Which[
        (value = TransitionPolarizability[k][StateLabel@state1] /. StateParameters@state2) =!= TransitionPolarizability[k][StateLabel@state1], value,
        (value = TransitionPolarizability[k][StateLabel@state2] /. StateParameters@state1) =!= TransitionPolarizability[k][StateLabel@state2], value,
        True, TransitionPolarizability[k][Sequence @@ Sort[StateLabel /@ {state1, state2}]]
    ]
];

GroundStateQ[state_AtomicState] := NaturalWidth@state === 0;
SetAttributes[GroundStateQ, Listable];

GroundStates[sys_?AtomicSystemQ] := Select[sys, GroundStateQ];

ExcitedStates[sys_?AtomicSystemQ] := Select[sys, !GroundStateQ[#]&];

SublevelMultiplicity[s_AtomicState] :=
    If[DefinedQ[s, F],
        (2F@s + 1),
        If[DefinedQ[s, J], (2 J@s + 1), 1]If[DefinedQ[s, NuclearSpin], (2 NuclearSpin@s + 1), 1]
    ];

FineStructureStatesQ[s1_AtomicState, s2_AtomicState] :=
    ListQ@StateLabel@s1 && ListQ@StateLabel@s2 && First@StateLabel@s1 === First@StateLabel@s2;

SublevelMultiplicity[sys_?AtomicSystemQ] :=
    Total[SublevelMultiplicity /@ (DeleteParameter[SelectState[sys, #], F]& /@ Union@StateLabel@sys)];

SelectStates[sys_?AtomicSystemQ, crit_, n_ : Infinity] := Extract[sys, StatePosition[sys, crit, n]];

SelectState[sys_?AtomicSystemQ, crit_] := First@SelectStates[sys, crit, 1];

DeleteStates[sys_?AtomicSystemQ, crit_] := Delete[sys, StatePosition[sys, crit]];

StateLabelQ[state_AtomicState, expr_] := StateLabel@state === expr;

StateLabelQ[sys_?AtomicSystemQ, expr_] := MemberQ[StateLabel@sys, expr];

StateEvaluate[state_AtomicState, expr_] := expr /. param : (StateParameterPattern[] | StateLabel) :> param@state;

StateEvaluate[sys_?AtomicSystemQ, expr_] := StateEvaluate[#, expr]& /@ sys;

StatePosition[sys_?AtomicSystemQ, crit_, n_ : Infinity] /; StateLabelQ[sys, crit] :=
    StatePosition[sys, StateLabel == crit, n];

StatePosition[sys_?AtomicSystemQ, crit_, n_ : Infinity]  /; !StateLabelQ[sys, crit] :=
    Position[sys, state_ /; StateEvaluate[state, crit], {1}, n, Heads -> False];

StateMatchQ[state_AtomicState, crit_] /; StateLabelQ[state, crit] := True;

StateMatchQ[state_AtomicState, crit_] := TrueQ@StateEvaluate[state, crit];

StateMatchQ[sys_?AtomicSystemQ, crit_] := StateMatchQ[#, crit]& /@ sys;

Submatrix[sys_, mat_, crit_Rule] :=
    Module[
        {pos = StatePosition[sys, #]& /@ crit},
        Extract[#, Last@pos] & /@ Extract[mat, First@pos]
    ];

Submatrix[sys_, mat_, crit_] :=
    Module[
        {pos = StatePosition[sys, crit]},
        Extract[#, pos] & /@ Extract[mat, pos]
    ];

TransitionEvaluate[state1_AtomicState, state2_AtomicState, crit1_ -> crit2_] :=
    StateMatchQ[state1, crit1] && StateMatchQ[state2, crit2];

TransitionEvaluate[state1_AtomicState, state2_AtomicState, crit : {__Rule}] :=
    And @@ (TransitionEvaluate[state1, state2, #] &) /@ crit;

TransitionEvaluate[state1_AtomicState, state2_AtomicState, crit_] :=
    crit /. {
        (param : (StateParameterPattern[] | StateLabel))' :> param@state2,
        param : (StateParameterPattern[] | StateLabel) :> param@state1
    };

TransitionEvaluate[sys1_?AtomicSystemQ, sys2_?AtomicSystemQ, crit_] :=
    Outer[
        Block[
            {br12Zero = PossibleZeroQ@BranchingRatio[#1][#2], br21Zero = PossibleZeroQ@BranchingRatio[#2][#1]},
            (!br12Zero || br21Zero) && TransitionEvaluate[#1, #2, crit]
                || (!br21Zero || br12Zero) && TransitionEvaluate[#2, #1, crit]
        ]&,
        sys1,
        sys2
    ];

MatrixElementPosition[sys_?AtomicSystemQ, crit1_ -> crit2_, n_ : Infinity] :=
    Module[{pos1, pos2},
        {pos1, pos2} = Flatten@StatePosition[sys, #, n] & /@ {crit1, crit2};
        Union[#, Reverse /@ #]&@Flatten[Outer[List, pos1, pos2], 1]
    ];
MatrixElementPosition[sys_?AtomicSystemQ, crit : {__Rule}] :=
    Union @@ (MatrixElementPosition[sys, #] &) /@ crit;
MatrixElementPosition[sys_?AtomicSystemQ, All | True] :=
    Tuples[Range@Length@sys, 2];
MatrixElementPosition[sys_?AtomicSystemQ, None | False] := {};
MatrixElementPosition[sys_?AtomicSystemQ, crit_, n_ : Infinity] :=
    Position[TransitionEvaluate[sys, sys, crit], True, {2}, n, Heads -> False];

MatrixElements[sys_?AtomicSystemQ, a_?OperatorQ, crit_] := Extract[a, MatrixElementPosition[sys, crit]];

MatrixElement[sys_?AtomicSystemQ, a_, {s1_AtomicState, s2_AtomicState}] :=
    Extract[a, Position[sys, #, {1}, 1, Heads -> False][[1, 1]]& /@ {s1, s2}];

(*MatrixElement[sys_?AtomicSystemQ, a_, crit_Rule] := First@Extract[a, MatrixElementPosition[sys, crit, 1]];*)

AtomicSystemType::difftypes = "All atomic states in the system are not of the same type, `1`, as the first state.";
AtomicSystemType::unrec = "The type of the atomic state `1` is not recognized.";

ValidQuantumNumberQ[s_AtomicState, j_] := DefinedQ[s, j] && HalfIntegerQ@j@s;

AtomicSystemType[s_AtomicState] :=
    Module[{noMessage = True, result},
        result = Check[
            Module[{fDefined, jDefined, mDefined},
                {fDefined, jDefined, mDefined} = ValidQuantumNumberQ[s, #]& /@ {F, J, M};
                Which[
                    fDefined && jDefined && mDefined,
                    "HyperfineZeeman",
                    fDefined && jDefined && !mDefined,
                    "Hyperfine",
                    !fDefined && jDefined && mDefined,
                    "Zeeman",
                    jDefined && !mDefined,
                    "Fine",
                    !jDefined && !mDefined,
                    "Toy",
                    True,
                    Message[AtomicSystemType::unrec, s]
                ]
            ],
            noMessage = False
        ];
        (AtomicSystemType[s] = result) /; noMessage
    ];

AtomicSystemType[sys_?AtomicSystemQ] :=
    Module[{noMessage = True, result},
        result = Check[
            With[{types = AtomicSystemType /@ sys},
                If[MatchQ[Hold@types, Hold@{_String..}],
                    If[MatchQ[types, {First@types..}],
                        First@types,
                        Message[AtomicSystemType::difftypes, First@types]
                    ]
                ]
            ],
            noMessage = False
        ];
        (AtomicSystemType[sys] = result) /; noMessage
    ];

SetAttributes[LevelLabel, Listable];
LevelLabel[s_AtomicState] := Switch[AtomicSystemType@s,
    "Toy" | "Zeeman" | "Fine",
    StateLabel@s,
    "HyperfineZeeman" | "Hyperfine",
    {StateLabel@s, F@s}
];

SetAttributes[SublevelLabel, Listable];
SublevelLabel[s_AtomicState] := Switch[AtomicSystemType@s,
    "Toy" | "Fine",
    StateLabel@s,
    "Zeeman",
    {StateLabel@s, M@s},
    "Hyperfine",
    {StateLabel@s, F@s},
    "HyperfineZeeman",
    {StateLabel@s, F@s, M@s}
];

HyperfineSublevels[s_AtomicState] := If[
    AtomicSystemType@s == "Fine" && DefinedQ[s, NuclearSpin] && HalfIntegerQ@NuclearSpin@s,
    Reverse@Table[AddParameter[s, F -> f], {f, Abs[NuclearSpin@s - J@s], NuclearSpin@s + J@s}],
    s
];

HyperfineSublevels[sys_?AtomicSystemQ] :=
    Flatten[HyperfineSublevels /@ sys];

ZeemanSublevels[s_AtomicState] :=
    Switch[AtomicSystemType@s,
        "Fine",
        Reverse@Table[AddParameter[s, M -> m], {m, -J@s, J@s}],
        "Hyperfine",
        Reverse@Table[AddParameter[s, M -> m], {m, -F@s, F@s}],
        _,
        s
    ];

ZeemanSublevels[sys_?AtomicSystemQ] :=
    Flatten[ZeemanSublevels /@ sys];

Sublevels[s_AtomicState] := ZeemanSublevels@HyperfineSublevels@s;

Sublevels[sys_?AtomicSystemQ] := Flatten[Sublevels /@ sys];

RemoveZeemanSublevels[sys_?AtomicSystemQ] := First /@ Split@DeleteParameter[sys, M];

CountZeemanSublevels[sys_?AtomicSystemQ] := Length /@ Split@DeleteParameter[sys, M];

ZeemanSystem[label_, j_] := ZeemanSublevels[AtomicState[label, J -> j]];

ZeemanSystem[j_] := ZeemanSystem[ToString@Unique[], j];

End[];

EndPackage[];
