(**************************************************************************

   	This file is part of the AtomicDensityMatrix package for Mathematica
    Copyright (C) 2004-2019 Simon M. Rochester <simon@rochesterscientific.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

**************************************************************************)


BeginPackage["AtomicDensityMatrix`DensityMatrix`", {"ProcessOptions`", "AtomicDensityMatrix`"}];

Begin["`Private`"];

Options[DensityMatrix] = {
    TimeDependence -> True,
    Representation -> "Zeeman",
    DMSymbol -> \[Rho],
    DMLabel -> None,
    ComplexExpandVariables -> False,
    TimeVariable -> t,
    AllowedElements -> All,
    NeglectedElements -> None,
    Subspace -> All
};

OptionProcessor[opt : TimeDependence | DMSymbol | DMLabel | ComplexExpandVariables | TimeVariable | AllowedElements | NeglectedElements | Subspace] :=
    (# /. Automatic :> OptionValue[DensityMatrix, opt] &);
OptionProcessor[DMLabel] := (# /. Automatic -> OptionValue[DensityMatrix, DMLabel] /. None | False -> Sequence[] &);
OptionProcessor[Representation] = Flatten@{# /. Automatic -> OptionValue[DensityMatrix, Representation]} /. PolarizationMoments -> "PolarizationMoments" &;

OptionValidator[TimeDependence] = True | False | All; (* All used in DMElementPattern *)
OptionValidator[ComplexExpandVariables] = True | False | Subscript;
OptionValidator[TimeVariable] = _Symbol | _?NumericQ;
OptionValidator[DMSymbol] = _Symbol;
OptionValidator[Representation] = {"Zeeman"} | {"PolarizationMoments", ___Rule};

Options[DMVariables] = Thread[Options[DensityMatrix][[All, 1]] -> Automatic];

Options[Ket] = FilterRules[Options@DMVariables, Except[DMSymbol | Representation | DMLabel]];

(Ket[sys_?AtomicSystemQ, symbol_, opts : OptionsPattern[]] := Module[{ket, complexExpandVariables, timeDependence, timeVariable},
    ket = Switch[complexExpandVariables,
        True,
        Re@# + I Im@#& /@ (Subscript[symbol, SublevelLabel@#]& /@ sys),
        Subscript,
        Subscript[symbol, Re, SublevelLabel@#] + I Subscript[symbol, Im, SublevelLabel@#]& /@ sys,
        _,
        Subscript[symbol, SublevelLabel@#]& /@ sys
    ];
    If[timeDependence, ket = ket /. x : Subscript[symbol, __] -> x[timeVariable]];
    Transpose@{ket}
]) // ProcessOptions;

Options[ExpandDMVariablesRule] = FilterRules[Options@DMVariables, Except[Representation | TimeVariable]];

(ExpandDMVariablesRule[sys_, opts : OptionsPattern[]] := Module[
    {sublevelLabels = SublevelLabel /@ sys, patt, rep, imp, elem, rept, impt, rep1, imp1, rept1, impt1, elemt, e, r, i, r1, i1, complexExpandVariables, dmSymbol, dmLabel},
    patt = DMElementPattern[ComplexExpandVariables -> False, TimeDependence -> All, FilterRules@opts];
    Switch[complexExpandVariables,
        True,
        (v : patt :> Switch[
            e = If[{Time} == {}, elem, elemt];
            Order[Position[sublevelLabels, State1, {1}], Position[sublevelLabels, State2, {1}]],
            -1, Re[e] - I Im[e],
            0, Re[v],
            1, Re[v] + I Im[v]
        ]) /. {
            elem -> Subscript[dmSymbol, dmLabel, State2, State1],
            elemt -> Derivative[TimeDerivative][Subscript[dmSymbol, dmLabel, State2, State1]][Time]
        },
        Subscript,
        (v : patt :> Switch[
            {r, i, r1, i1} = If[{Time} == {}, {rep, imp, rep1, imp1}, {rept, impt, rept1, impt1}];
            Order[Position[sublevelLabels, State1, {1}], Position[sublevelLabels, State2, {1}]],
            -1, r1 - I i1,
            0, r,
            1, r + I i
        ]) /. {
            rep -> Subscript[dmSymbol, Re, dmLabel, State1, State2],
            imp -> Subscript[dmSymbol, Im, dmLabel, State1, State2],
            rept -> Derivative[TimeDerivative][Subscript[dmSymbol, Re, dmLabel, State1, State2]][Time],
            impt -> Derivative[TimeDerivative][Subscript[dmSymbol, Im, dmLabel, State1, State2]][Time],
            rep1 -> Subscript[dmSymbol, Re, dmLabel, State2, State1],
            imp1 -> Subscript[dmSymbol, Im, dmLabel, State2, State1],
            rept1 -> Derivative[TimeDerivative][Subscript[dmSymbol, Re, dmLabel, State2, State1]][Time],
            impt1 -> Derivative[TimeDerivative][Subscript[dmSymbol, Im, dmLabel, State2, State1]][Time]
        },
        _,
        {}
    ]
]) // ProcessOptions;

(DensityMatrix[system_?AtomicSystemQ, opts : OptionsPattern[]] := Module[
    {subspace, representation, timeDependence, timeVariable, dmLabel, dmSymbol, complexExpandVariables,
        allowedElements, neglectedElements, labels, pos, elements, sys},
    sys = If[subspace =!= All, SelectStates[sys, subspace], system];
    Switch[representation,
        {"Zeeman"},
        labels = SublevelLabel@sys;
        pos = Complement @@ (MatrixElementPosition[sys, #]&) /@ {allowedElements, neglectedElements};
        If[pos === {},
            ZeroMatrix@Length@sys,
            elements = MapThread[Subscript[dmSymbol, dmLabel, ##]&, labels[[#]]& /@ Transpose@pos];
            If[timeDependence, elements = Through@elements@timeVariable];
            If[MatchQ[complexExpandVariables, True | Subscript],
                elements = elements /. ExpandDMVariablesRule[sys, FilterRules@opts]
            ];
          Normal@SparseArray[pos -> elements, {Length@sys, Length@sys}]
        ],
        {"PolarizationMoments", ___},
        PolarizationMoments[sys, Rest@representation, FilterRules@opts]
    ]
]) // ProcessOptions;

DensityMatrix[as_?AtomicSystemQ, es : {_ExperimentalRegion..}, opts : OptionsPattern[]] :=
    DensityMatrix[as, DMLabel -> RegionLabel@#, opts]& /@ es;

DMSubstitute[sys_, mat_] := Thread[Flatten@DensityMatrix[sys] -> Flatten@mat];

AllowedPositions[dm_?OperatorQ] := Position[dm, Except@0, {2}, Heads -> False];

AllowedUpperTrianglePositions[dm_?OperatorQ] :=
    Reverse /@ Sort[Reverse /@ Select[AllowedPositions@dm, OrderedQ]]; (* This sorting is to match the legacy sorting of UpperTriangle. Might want to just sort in a more natural way. *)

AllowedPositions[dm_?TensorDecompositionMatrixQ] := Position[dm, Except@0, {5}, Heads -> False];

AllowedUpperTrianglePositions[dm_?ContravariantDecompositionMatrixQ] := Module[{rank = Rank@dm},
    Select[AllowedPositions@dm, #[[1]] < #[[2]] || #[[1]] == #[[2]] && #[[4]] <= Extract[rank, Take[#, 3]] + 1 &]
];

(DMVariables[sys_?AtomicSystemQ, opts : OptionsPattern[]] := Module[
    {complexExpandVariables, dm, posFunc},
    dm = DensityMatrix[sys, FilterRules@opts];
    posFunc = Switch[complexExpandVariables,
        True | Subscript,
        AllowedUpperTrianglePositions,
        False,
        AllowedPositions
    ];
    Flatten[Extract[dm, posFunc@dm] /. x_ + I y_ -> {x, y}]
]) // ProcessOptions;

DMVariables[as_?AtomicSystemQ, es : {_ExperimentalRegion..}, opts : OptionsPattern[]] :=
    Flatten[DMVariables[as, DMLabel -> RegionLabel@#, opts]& /@ es];

Quiet[LevelLabelPattern[State_, StateLabel_, F_, M_] :=
    State : {StateLabel_, F_, M_} | {StateLabel_, M_} | StateLabel_?(!ListQ[#]&), {RuleDelayed::"rhs"}];

LevelLabelPattern1[] := LevelLabelPattern[State1, StateLabel1, F1, M1];

LevelLabelPattern2[] := LevelLabelPattern[State2, StateLabel2, F2, M2];

Options[DMElementPattern] = FilterRules[Options@DMVariables, Except[Representation | DMLabel | TimeVariable]];

(DMElementPattern[opts : OptionsPattern[]] := Module[{newopts, dmSymbol, complexExpandVariables, timeDependence},
    Switch[{complexExpandVariables, timeDependence},
        {Subscript, False}, dmElementPartPattern@dmSymbol,
        {Subscript, True}, dmElementPartTDPattern@dmSymbol,
        {Subscript, All}, dmElementPartPattern@dmSymbol | dmElementPartTDPattern@dmSymbol,
        {_, False}, dmElementPattern@dmSymbol,
        {_, True}, dmElementTDPattern@dmSymbol,
        {_, All}, dmElementPattern@dmSymbol | dmElementTDPattern@dmSymbol
    ]
]) // ProcessOptions;

dmElementPattern[r_] = (
    Subscript[r, DMLabel_, LevelLabelPattern1[], LevelLabelPattern2[]] |
        Subscript[r, LevelLabelPattern1[], LevelLabelPattern2[]]
) /; (DMLabel =!= Re && DMLabel =!= Im);

dmElementTDPattern[r_] = (
    Derivative[TimeDerivative_][Subscript[r, DMLabel_, LevelLabelPattern1[], LevelLabelPattern2[]]][Time_] |
        Subscript[r, DMLabel_, LevelLabelPattern1[], LevelLabelPattern2[]][Time_] |
        Derivative[TimeDerivative_][Subscript[r, LevelLabelPattern1[], LevelLabelPattern2[]]][Time_] |
        Subscript[r, LevelLabelPattern1[], LevelLabelPattern2[]][Time_]
) /; (DMLabel =!= Re && DMLabel =!= Im);

dmElementPartPattern[r_] = (
    Subscript[r, ReImPart : Re | Im, DMLabel_, LevelLabelPattern1[], LevelLabelPattern2[]] |
        Subscript[r, ReImPart : Re | Im, LevelLabelPattern1[], LevelLabelPattern2[]]
) /; (DMLabel =!= Re && DMLabel =!= Im);

dmElementPartTDPattern[r_] = (
    Derivative[TimeDerivative_][Subscript[r, ReImPart : Re | Im, DMLabel_, LevelLabelPattern1[], LevelLabelPattern2[]]][Time_] |
        Subscript[r, ReImPart : Re | Im, DMLabel_, LevelLabelPattern1[], LevelLabelPattern2[]][Time_] |
        Derivative[TimeDerivative_][Subscript[r, ReImPart : Re | Im, LevelLabelPattern1[], LevelLabelPattern2[]]][Time_] |
        Subscript[r, ReImPart : Re | Im, LevelLabelPattern1[], LevelLabelPattern2[]][Time_]
) /; (DMLabel =!= Re && DMLabel =!= Im);

Options[ExprToReIm] == FilterRules[Options@DMVariables, Except[ComplexExpandVariables | Representation | DMLabel | TimeVariable]];

ExprToReIm[expr_List, opts : OptionsPattern[]] := Flatten[ExprToReIm[#, opts]& /@ expr];

(ExprToReIm[expr_, opts : OptionsPattern[]] :=
    DeleteCases[
        Chop@Refine[
            ComplexExpand[
                Flatten@Through[{Re, Im}@expr] /. {(f : Re | Im)[x_ == y_] :> f[x] == f[y]},
                DMElementPattern[ComplexExpandVariables -> False, FilterRules@opts, TimeDependence -> All],
                TargetFunctions -> {Re, Im}
            ],
            DMElementPattern[ComplexExpandVariables -> Subscript, FilterRules@opts, TimeDependence -> All] \[Element] Reals
        ] /. {
            Re[x : DMElementPattern[ComplexExpandVariables -> Subscript, FilterRules@opts, TimeDependence -> All]] -> x,
            Im[x : DMElementPattern[ComplexExpandVariables -> Subscript, FilterRules@opts, TimeDependence -> All]] -> 0
        },
        True | 0
    ]) // ProcessOptions;

ExpandExp[expr_] := expr /. y : Exp[_] :> ExpToTrig[y];

MyExpand[expr_] := expr //. x_ y_Plus :> Distribute[x y];

SetAttributes[ERI, Listable];

ERI[x_ == y_] := MapThread[Equal, ERI@{x, y}];

ERI[expr_] := Total@Flatten@# & /@ Last@Reap[ERISow[expr /. y : Exp[_] :> ExpToTrig[y] //. x_ y_Plus :> Distribute[x y]], {Re, Im}];

ERISow[x_Plus] := Scan[ERISow, x];

ERISow[x_] :=
    If[
        FreeQ[x, Complex],
        Sow[x, Re],
        If[
            MatchQ[x, Complex[0, a_] b_. /; FreeQ[b, Complex]],
            Sow[x / I, Im],
            Sow[ComplexExpand[#[x], TargetFunctions -> {Re, Im}], #] /@ {Re, Im};
        ]
    ];


PopulatedDM[sys_?AtomicSystemQ, crit_] := PopulatedDM[sys, crit, 1];
PopulatedDM[sys_?AtomicSystemQ, crit_, weight_] := DiagonalMatrix@Normalize[Boole@StateMatchQ[sys, crit] StateEvaluate[sys, weight], Total];
PopulatedDM[j_?HalfIntegerQ, crit_] := PopulatedDM[ZeemanSystem[j], crit];

End[];

EndPackage[];
