(**************************************************************************

   	This file is part of the AtomicDensityMatrix package for Mathematica
    Copyright (C) 2004-2019 Simon M. Rochester <simon@rochesterscientific.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

**************************************************************************)


BeginPackage["AtomicDensityMatrix`LiouvilleEquation`", {"ProcessOptions`", "AtomicDensityMatrix`"}];

Begin["`Private`"];

Options[LiouvilleEquation] = Options[DMVariables];

fixedD[expr_, x_] := D[expr, x] /. {Re'[e_] :> Re[D[e, x]] / D[e, x], Im'[e_] :> Im[D[e, x]] / D[e, x]};

pruneEquations[dm_, complexExpandVariables_, eqs_] :=
    Switch[complexExpandVariables,
        False,
        Extract[eqs, AllowedPositions@dm],
        True | Subscript,
        ERI@Extract[eqs, AllowedUpperTrianglePositions@dm]
    ] // DeleteCases[Flatten@#, True]&;

(LiouvilleEquation[sys_?AtomicSystemQ, h_?OperatorQ, relax_?OperatorQ, repop_?OperatorQ, opts : OptionsPattern[]] :=
    Module[{representation, timeVariable, complexExpandVariables, allowedElements, neglectedElements,  lhs, rhs,  h1,
        relax1, repop1, dm, neglectedVariables},
        Switch[representation,
            {"PolarizationMoments", ___Rule},
            LiouvilleEquation[sys, ##, opts]& @@ (Decomposition[sys, #]&) /@ {h, relax, repop},
            {"Zeeman"},
            {h1, relax1, repop1} = {h, relax, repop} /. ExpandDMVariablesRule[sys, FilterRules@opts];
            dm = DensityMatrix[sys, FilterRules@opts];
            neglectedVariables = DMVariables[sys, AllowedElements -> neglectedElements, NeglectedElements -> allowedElements, FilterRules@opts];
            lhs = fixedD[dm, timeVariable];
            rhs = -I Commutator[h1, dm] - Anticommutator[dm, relax1] / 2 + repop1;
            pruneEquations[
                dm,
                complexExpandVariables,
                MapThread[Equal, {lhs, rhs}, 2] /. Dispatch@Thread[neglectedVariables -> 0]
            ]
        ]
    ]) // ProcessOptions;

LiouvilleEquation[sys_?AtomicSystemQ, h_?OperatorQ, relax_?OperatorQ, opts : OptionsPattern[]] :=
    LiouvilleEquation[sys, h, relax, ZeroMatrix@Length@sys, opts];

LiouvilleEquation[sys_?AtomicSystemQ, h_?OperatorQ, opts : OptionsPattern[]] :=
    LiouvilleEquation[sys, h, ZeroMatrix@Length@sys, ZeroMatrix@Length@sys, opts];

LiouvilleEquation[
    as_?AtomicSystemQ,
    es : {_ExperimentalRegion..},
    h : {_?OperatorQ..},
    relax : {_?OperatorQ..},
    repop : {_?OperatorQ..},
    opts : OptionsPattern[]] :=
    Flatten[
        MapThread[
            LiouvilleEquation[as, #2, #3, #4, DMLabel -> RegionLabel@#1, opts]&,
            {es, h, relax, repop}
        ],
        1
    ];

(LiouvilleEquation[sys_?AtomicSystemQ, h_?ContravariantDecompositionMatrixQ,
    relax_?ContravariantDecompositionMatrixQ, repop_?ContravariantDecompositionMatrixQ, opts : OptionsPattern[]] :=
    Module[{representation, timeVariable, allowedElements, neglectedElements, complexExpandVariables, dm, eqs,
        neglectedVariables, n, f, component, levels, rank},
        Switch[representation,
            {"Zeeman"},
            LiouvilleEquation[sys, ##, opts]& @@ (Recomposition[sys, #]&) /@ {h, relax, repop},
            {"PolarizationMoments", ___Rule},
            levels = RemoveZeemanSublevels@sys;
            rank = Rank@h;
            dm = DensityMatrix[sys, FilterRules@opts];
            neglectedVariables = DMVariables[sys, AllowedElements -> neglectedElements, NeglectedElements -> allowedElements, FilterRules@opts];
            n = Length@levels;
            f = F@levels;
            component = Function[{op, i1, i2, k, q}, op[[i1, i2, k - Abs[f[[i1]] - f[[i2]]] + 1, -q + k + 1, 1]]];
            eqs = Table[
                Transpose@List@Table[
                    D[component[dm, i1, i2, k, q], timeVariable] ==
                        Sum[
                            With[{q2 = q - q1},
                                -I (-1)^(f[[i1]] + f[[i2]] + k) Sqrt[(2 k1 + 1) (2 k2 + 1)] CG[{k1, q1}, {k2, q2}, {k, q}] *
                                    SJS[{k1, k2, k}, {f[[i2]], f[[i1]], f[[i3]]}] *
                                        ((component[h, i1, i3, k1, q1] - I / 2 component[relax, i1, i3, k1, q1]) component[dm, i3, i2, k2, q2]
                                            - component[dm, i1, i3, k1, q1] (component[h, i3, i2, k2, q2] + I / 2 component[relax, i3, i2, k2, q2]))
                            ],
                            {i3, n},
                            {k1, rank[[i1, i3, 1]], rank[[i1, i3, -1]]},
                            {k2, rank[[i3, i2, 1]], rank[[i3, i2, -1]]},
                            {q1, Min[k1, q + k2], Max[-k1, q - k2], -1}
                        ] + component[repop, i1, i2, k, q],
                    {q, k, -k, -1}
                ],
                {i1, n},
                {i2, n},
                {k, rank[[i1, i2, 1]], rank[[i1, i2, -1]]}
            ];
            pruneEquations[dm, complexExpandVariables, eqs /. Dispatch@Thread[neglectedVariables -> 0]]
        ]
    ]) // ProcessOptions;

Options[InitialConditions] = FilterRules[Options@DMVariables, Except@TimeVariable];

(InitialConditions[sys_?AtomicSystemQ, InitialDM_?OperatorQ, t0_, opts : OptionsPattern[]]:=
    Module[{lhs, complexExpandVariables},
        lhs = DensityMatrix[sys, TimeVariable -> t0, FilterRules@opts];
        pruneEquations[
            lhs,
            complexExpandVariables,
            MapThread[Equal, {lhs, InitialDM}, 2]
        ]
    ]) // ProcessOptions;

InitialConditions[as_?AtomicSystemQ, es : {_ExperimentalRegion..}, InitialDM : {_?OperatorQ..}, t0_, opts : OptionsPattern[]] :=
    Flatten[MapThread[InitialConditions[as, #2, t0, DMLabel -> RegionLabel@#1, opts]&, {es, InitialDM}], 1];

End[];

EndPackage[];
