(**************************************************************************

   	This file is part of the AtomicDensityMatrix package for Mathematica
    Copyright (C) 2004-2019 Simon M. Rochester <simon@rochesterscientific.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

**************************************************************************)

BeginPackage["AtomicDensityMatrix`MatrixRSolve`", {"AtomicDensityMatrix`"}];

If[$VersionNumber < 7., Get["FourierSeries`"]];

EquationsToMatrixRecursion::oddreq = "An odd number of harmonics must be specified. Using `1` harmonics.";

Begin["`Private`"];

FactorsContaining[expr_SparseArray, x_] := FactorsContaining[expr["NonzeroValues"], x];

FactorsContaining[expr_, x_] := Select[Variables[expr /. Power -> power] /. power -> Power, ! FreeQ[#, x] &];

ExponentialFactors[t_, w_, l_Integer] := Delete[Table[factor[t, w, k], {k, -l, l}], l + 1];

ExponentialFactorReplacement[t_, w_, n_Integer] :=
    Thread[Delete[Exp[I t w Range[-n, n] + a_.], n + 1] -> Exp[a] ExponentialFactors[t, w, n]];

FourierSeriesAtFrequency[list_List, t_, w_, n_Integer] :=
    FourierSeriesAtFrequency[#, t, w, n] & /@ list;

fourierParameters = If[$VersionNumber >= 7., {1, 1}, {Pi, 1 / (2 Pi)}];

FourierSeriesAtFrequency[expr_, t_, w_, n_Integer] :=
    FourierSeriesAtFrequency[expr, t, w, n] =
        FourierSeries[expr /. t -> t / w, t, n, FourierParameters -> fourierParameters] /. {t -> w t};

BMatrices[blochMatrix_, t_, w_, l_Integer] :=
    Module[{coefficients, expandedOscillatingFactors, oscillatingFactors},
        oscillatingFactors = FactorsContaining[blochMatrix, t];
        expandedOscillatingFactors = FourierSeriesAtFrequency[oscillatingFactors, t, w, l] /.
            ExponentialFactorReplacement[t, w, l];
        coefficients = CoefficientArrays[
            blochMatrix /. Thread[oscillatingFactors -> expandedOscillatingFactors],
            ExponentialFactors[t, w, l]
        ];
        Insert[Transpose[coefficients[[2]], {2, 3, 1}], coefficients[[1]], l + 1]
    ];

BHatMatrices[bMatrices_, l_Integer] :=
    SparseArray /@ ArrayFlatten /@ {
        PadLeft@Table[bMatrices[[k ;; l]], {k, l}],
        Reverse@Table[bMatrices[[k + 1 ;; k + l]], {k, l}],
        PadRight@Table[bMatrices[[-l ;; -1]], {l}]
    };

CHatVector[cVector_, l_Integer] := PadLeft[cVector, l Length@cVector, 0, (l - 1) / 2 Length@cVector];

EquationsToMatrixRecursion[eqs_, vars_, w_, t_, l_?OddQ] :=
    Module[{dvars, cVector, blochMatrix},
        dvars = D[vars, t];
        {cVector, blochMatrix} = CoefficientArrays[dvars /. First@Solve[eqs, dvars], vars];
        {BHatMatrices[BMatrices[Normal@blochMatrix, t, w, l], l], CHatVector[cVector, l], vars, w, t, l}
    ];

EquationsToMatrixRecursion[eqs_, vars_, w_, t_, l_?EvenQ] := (
    Message[EquationsToMatrixRecursion::oddreq, l + 1];
    EquationsToMatrixRecursion[eqs, vars, w, t, l + 1]
);

DHatMatrix[m_Integer, l_Integer, nvars_, w_] := I * w * DiagonalMatrix@SparseArray@Flatten@Transpose@ConstantArray[Range[m l + -(l - 1) / 2, m l + (l - 1) / 2], nvars];

SIterator[recursion : {{bHatMinus_, bHatZero_, bHatPlus_}, cHatVector_, vars_, w_, t_, l_Integer}, {sPlus_, mPlus_}] :=
    Module[
        {m = mPlus - 1},
        {-SparseArray@LinearSolve[bHatZero - DHatMatrix[m, l, Length@vars, w] + bHatPlus.sPlus, bHatMinus], m}
    ];

SList[recursion : {{bHatMinus_, bHatZero_, bHatPlus_}, cHatVector_, vars_, w_, t_, l_Integer}, m_Integer] :=
    Reverse@NestList[
        SIterator[recursion, #] &,
        {ZeroMatrix[Length@bHatZero, SparseArray], m},
        m - 1
    ][[All, 1]];

TildeOp[mat_, nvars_Integer] :=
    SparseArray[Conjugate@ArrayFlatten@Reverse[Partition[mat, {nvars, nvars}], {1, 2}]];

AHatZero[recursion : {{bHatMinus_, bHatZero_, bHatPlus_}, cHatVector_, vars_, w_, t_, l_Integer}, s1_] :=
    -LinearSolve[
        bHatMinus.TildeOp[s1, Length@vars] + bHatZero - DHatMatrix[0, l, Length@vars, w] + bHatPlus.s1,
        cHatVector
    ];

AHatIterator[{aHatMinus_, nMinus_}, sList_] := {sList[[nMinus + 1]].aHatMinus, nMinus + 1};

AHatList[recursion : {{bHatMinus_, bHatZero_, bHatPlus_}, cHatVector_, vars_, w_, t_, l_Integer}, sList_, n2_Integer] :=
    Take[
        Drop[
            Partition[
                Flatten@NestList[
                    AHatIterator[#, sList] &,
                    {AHatZero[recursion, First@sList], 0},
                    Ceiling[n2 / l]
                ][[All, 1]],
                Length@vars
            ],
            (l - 1) / 2
        ],
        n2 + 1
    ];

RecursionSolutionToReplacements[sol_, vars_, t_] :=
    Flatten[MapIndexed[MapThread[Rule, {vars /. t -> #2[[1]] - 1, #1}] &, sol]];

MatrixRSolve[recursion : {{bHatMinus_, bHatZero_, bHatPlus_}, cHatVector_, vars_, w_, t_, l_Integer}, n1_Integer, n2_Integer] /; n1 >= n2 :=
    RecursionSolutionToReplacements[AHatList[recursion, SList[recursion, n1], n2], vars, t];

MatrixRSolve[recursion_, n1_Integer, n2_Integer, parameters_] /; n1 >= n2 :=
    MatrixRSolve[SparseReplaceAll[recursion, parameters], n1, n2];

FourierExpandVariables[vars_List, w_, t_, nmax_Integer] :=
    ComplexExpand[
        Sum[(vars /. t -> n) E^(I n w t), {n, -nmax, nmax}] /.
            (((# /. t -> n_) /; n < 0 :> Conjugate[# /. t -> -n]) & /@ vars),
        vars /. t -> _,
        TargetFunctions -> {Re, Im}
    ] /. ((Im[# /. t -> 0] -> 0) & /@ vars);

FourierExpandVariables[vars_, w_, t_, nmax_Integer] :=
    First@FourierExpandVariables[{vars}, w, t, nmax];

FourierExpandVariables[expr_, vars_List, w_, t_, nmax_Integer] :=
    expr /. MapThread[Rule, {vars, FourierExpandVariables[vars, w, t, nmax]}];

FourierExpandVariables[expr_, vars_, w_, t_, nmax_Integer] :=
    FourierExpandVariables[expr, {vars}, w, t, nmax];

FourierExpandVariables[expr_, vars_, w_, t_, nvar_] :=
    TrigToExp[expr] //.
        {(a_ + b_)c_ -> a c + b c} /.
            Dispatch@Flatten[
                {
                    # -> FE[# /. t -> nvar, w, nvar],
                    D[#, {t, j_}] -> (I w nvar)^j FE[# /. t -> nvar, w, nvar]
                }& /@ vars
            ];

IntegrateFourierTerms[polys_List, vars_, w_, t_, nvar_] :=
    IntegrateFourierTerms[#, vars, w, t, nvar]& /@ polys;

IntegrateFourierTerms[poly_, vars_, w_, t_, nvar_] :=
    Module[
        {FEvars = Thread@FE[vars /. t -> nvar, w, nvar], arrays, exp},
        arrays = Normal@CoefficientArrays[poly, FEvars] /. Exp[x_] :> Distribute[exp[ExpandAll[x]], Plus, exp, Times, exp];
        arrays[[1]] KroneckerDelta[nvar, 0] +
            If[Length[arrays] == 2,
                Expand[arrays[[2]].FEvars] /. {
                    exp[Complex[0, j_] t w] FE[a_, w, nvar] :> (a /. nvar -> nvar - j),
                    FE[a_, w, nvar] -> a
                } /. exp -> Exp,
                0
            ] /; (PolynomialQ[poly, FEvars] && Max@Exponent[poly, FEvars] <= 1)
    ];

FourierExpandAndIntegrate[polys_, vars_, w_, t_] :=
    Module[{n},
        IntegrateFourierTerms[FourierExpandVariables[polys, vars, w, t, n], vars, w, t, n] /.
            n -> 0 /.
                (((# /. t -> 0) -> Re[# /. t -> 0])& /@ vars) /.
                    (((# /. t -> (m_ /; m < 0)) -> Conjugate[# /. t -> -m])& /@ vars)
    ];

SparseReplaceAll[s_SparseArray, rule_] := With[
    {elem = s["NonzeroValues"] /. rule, def = s["Background"] /. rule},
    Replace[s, Verbatim[SparseArray][a__, _, {b__, _}] :> SparseArray[a, def, {b, elem}]]
];

SparseReplaceAll[s_List, rule_] := SparseReplaceAll[#, rule] & /@ s;

SparseReplaceAll[s_, rule_] := s /. rule;

End[];

EndPackage[];
