(**************************************************************************

   	This file is part of the AtomicDensityMatrix package for Mathematica
    Copyright (C) 2004-2019 Simon M. Rochester <simon@rochesterscientific.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

**************************************************************************)

BeginPackage["AtomicDensityMatrix`PlotAndFit`", {"ProcessOptions`", "AtomicDensityMatrix`"}];

If[$VersionNumber < 7., Needs["NonlinearRegression`"]];

Begin["`Private`"];

Options[PlotAndFit] = Union[
    {
        DoFit -> True,
        ShowPlot -> True,
        HoldFixed -> {},
        Joined -> False,
        PlotStyle -> {Hue[0.67`, 0.6`, 0.6`], Red},
        Frame -> True,
        Interactive -> False,
        PlotRange -> All,
        ImageSize -> Small
    },
    DeleteCases[
        Flatten@{Options@Plot, Options@ListPlot, If[$VersionNumber < 7., Options@NonlinearRegress, Options@NonlinearModelFit]},
        WorkingPrecision | Joined | PlotStyle | Frame | PlotRange | ImageSize -> _
    ]
];

makeVector[expr_] := If[VectorQ@expr, expr, {expr}];

makeMatrix[expr_] := If[MatrixQ@expr, expr, {expr}];

ensureList[expr_] := If[ListQ@expr, expr, {expr}];

OptionProcessor[HoldFixed] = makeVector;

(PlotAndFit[data_, function_, params_, var_, opts : OptionsPattern[]] := Module[
    {interactive, method, holdFixed, doFit, showPlot, plotStyle, fit, fixpos, params1, params2, fixrules},
    (
        params1 = makeMatrix@params;
        fixpos = Position[params1, Alternatives @@ ({#, _}& /@ holdFixed), {1}];
        params2 = Delete[params1, fixpos];
        fixrules = Rule @@@ Extract[params1, fixpos];
        fit = If[doFit,
            If[$VersionNumber < 7.,
                BestFitParameters /. NonlinearRegress[data, function /. fixrules, params2, var, RegressionReport -> {BestFitParameters}],
                NonlinearModelFit[data, function /. fixrules, params2, var, Method -> Automatic, FilterRules@opts]["BestFitParameters"]
            ],
            Rule @@@ params2
        ];
        If[showPlot,
            Print@Show[
                ListPlot[data, PlotStyle -> First@ensureList@plotStyle, FilterRules@opts],
                Plot[
                    function /. fixrules /. fit,
                    {var, Min@data[[All, 1]], Max@data[[All, 1]]},
                    PlotStyle -> Last@ensureList@plotStyle, Method -> Automatic, Evaluate@FilterRules@opts
                ],
                Evaluate@FilterRules[{opts}, {PlotRange}]
            ]
        ];
        Transpose@{params1[[All, 1]], params1[[All, 1]] /. fixrules /. fit}
    ) /; !interactive || MemberQ[method, "Noninteractive"]
]) // ProcessOptions;

FitAndErrorNR[data_, func_, params_, x_] := Module[{names = params[[All, 1]], fitvals, errors, fit},
    {fitvals, errors} = If[$VersionNumber < 7.,
        fit = NonlinearRegress[data, func, params, x];
        {names /. (BestFitParameters /. fit), (ParameterCITable /. fit)[[1, All, 2]]},
        fit = NonlinearModelFit[data, func, params, x];
        {names /. fit@"BestFitParameters", fit@"ParameterErrors"}
    ];
    Thread[names -> Transpose@{fitvals, errors}]
];

(PlotAndFit[data_, func_, params_, x_, opts : OptionsPattern[]] := Module[{method, interactive, showPlot, plotStyle, output, plot},
    (
        DynamicModule[
            {params1, fitq, names, ivalues, fvalues, evalues, nv = Length[params],
                doFit = False, fitDone = False, setvals, newparams, setvals1},
            params1 = ReplacePart[Table[{0, 1, 1}, {nv}], Flatten@MapIndexed[#2 -> #1&, ensureList /@ params, {2}]];
            names = params1[[All, 1]];
            ivalues = params1[[All, 2]];
            fitq = (# =!= 0)& /@ params1[[All, 3]];
            fvalues = ivalues;
            evalues = 0 fvalues;
            output = DialogInput[
                DialogNotebook[
                    Panel[
                        Grid[
                            {{
                                Column[{
                                    Grid[
                                        Prepend[
                                            Table[
                                                With[{i = i}, {
                                                    names[[i]],
                                                    InputField[Dynamic@ivalues[[i]], FieldSize -> Small],
                                                    Checkbox@Dynamic@fitq[[i]],
                                                    InputField[Dynamic@fvalues[[i]], FieldSize -> Small, Enabled -> False],
                                                    InputField[Dynamic@evalues[[i]], FieldSize -> Small, Enabled -> False]
                                                }],
                                                {i, nv}
                                            ],
                                            {"", "Initial value", "Fit?", "Fit value", "Error"}
                                        ]],
                                    Row[{
                                        Button["Plot", fitDone = False],
                                        Button["Do Fit", doFit = True],
                                        DefaultButton["Done", DialogReturn@{plot, Transpose@{names, fvalues}}]
                                    }]
                                }],
                                Dynamic[
                                    If[doFit,
                                        setvals = Pick[Thread[names -> ivalues], fitq, False];
                                        newparams = Pick[Transpose@{names, ivalues}, fitq];
                                        setvals1 = Thread[names -> Thread@{ivalues, 0}];
                                        {fvalues, evalues} = Transpose[names /. FitAndErrorNR[data, func /. setvals, newparams, x] /. setvals1];
                                        doFit = False;
                                        fitDone = True
                                    ];
                                    plot = Show[
                                        ListPlot[data, PlotStyle -> First@ensureList@plotStyle, FilterRules@opts],
                                        Plot[
                                            func /. Thread[names -> If[fitDone, fvalues, ivalues]],
                                            {x, Min@data[[All, 1]], Max@data[[All, 1]]},
                                            PlotStyle -> Last@ensureList@plotStyle, Method -> Automatic, Evaluate@FilterRules@opts
                                        ],
                                        Evaluate@FilterRules[{opts}, {PlotRange}]
                                    ];
                                    Show[plot, ImageSize -> Medium]
                                ]
                            }},
                            Alignment -> Top
                        ]
                    ]
                ]
            ]
        ];
        If[showPlot, Print@First@output];
        Last@output
    ) /; interactive || !MemberQ[method, "Noninteractive"]
]) // ProcessOptions;

End[];

EndPackage[];
