(**************************************************************************

   	This file is part of the AtomicDensityMatrix package for Mathematica
    Copyright (C) 2004-2019 Simon M. Rochester <simon@rochesterscientific.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

**************************************************************************)

BeginPackage["AtomicDensityMatrix`TensorAlgebra`", {"ProcessOptions`", "AtomicDensityMatrix`"}];

TensorType::unrec = "The expression `1` does not have a recognized tensor structure.";

Dual::wrongtype = "Cannot take the dual of tensor of type `1`.";

ToContravariant::wrongtype = "Tensor of type `1` cannot be converted to contravariant form.";

ToCovariant::wrongtype = "Tensor of type `1` cannot be converted to covariant form.";

ToCartesian::wrongtype = "Tensor of type `1` cannot be converted to Cartesian form.";

Component::wrongtype = "Tensor of type `1` does not have a single component.";
Component::noexist = "The component `2` of tensor of type `1` does not exist.";
Component::noexistn = "The component of tensor of type `1` specified by the sequence `2` does not exist.";
Component::wrongspec = "`2` is not an allowed component specification for tensor of type `1`.";
Component::extraindices = "`3` indices specified for tensor of type `1`, which can take at most `2`.";

LowerTriangle::wrongtype = "Cannot take lower triangle of tensor of type `1`.";

UpperTriangle::wrongtype = "Cannot take upper triangle of tensor of type `1`.";

MatrixMultiply::nomult = "Tensors of type `1` and `2` cannot be multiplied in this manner.";
MatrixMultiply::dotsh = "Tensors of type `1` and `2` have incompatible dimensions.";

Begin["`Private`"];

StretchedState[j_?HalfIntegerQ] := ECT@List@Table[KroneckerDelta[1, i], {i, 2j + 1}];

(* Determining tensor structure *)

$TensorTypeStrings = {"Scalar", "Cartesian", "Contravariant", "Covariant", "Operator", "OneByOne", "TensorDecomposition", "TensorDecompositionMatrix"};

TensorTypePattern[] = {(Alternatives @@ $TensorTypeStrings)..};

TypedTensorPattern[] = {_, TensorTypePattern[]};

TensorTypeQ[a_] := MatchQ[a, TensorTypePattern[]];

SetAttributes[AddTensorTypeToCall, HoldFirst];
AddTensorTypeToCall[def_] := Module[{lhs, argPatts, tensorPos, newDef, typeVar, args, newArgPatts, newArgs, f,
    typePatternPattern, firstDef, condition = True},
    If[Extract[Hold@def, {1, 0}, Hold] === Hold@CompoundExpression,
        firstDef = Extract[Hold@def, {1, 1}, Hold],
        firstDef = Hold@def
    ];
    lhs = Extract[firstDef, {1, 1}, Hold];
    If[lhs[[1, 0]] === Condition,
        condition = Extract[lhs, {1, 2}, Hold];
        lhs = Extract[lhs, {1, 1}, Hold]
    ];
    argPatts = ReleaseHold[List @@@ lhs];
    typePatternPattern =
        (p_Pattern /; First@p === type) |
            (p : (_PatternTest | _Pattern) /; Last@p === TensorTypeQ) |
                _?TensorTypeQ |
                    HoldPattern[TensorTypePattern[]];
    tensorPos = Position[argPatts, {_, typePatternPattern}, {1}, 1, Heads -> False][[1, 1]];
    args = argPatts //. p_PatternTest | p_Pattern | p_Optional :> First@p;
    With[{tensorName = argPatts[[tensorPos, 1, 1]]},
        newArgPatts = EvaluateHold@ReplacePart[argPatts, tensorPos -> tensorName : Except@TypedTensorPattern[]]
    ];
    newArgs = EvaluateHold@ReplacePart[args, {tensorPos, 2} -> type$];
    f = lhs[[1, 0]];
    typeVar = args[[tensorPos, 2]];
    With[{
        newCondition = condition /. typeVar -> type$,
        newCall = HoldApply[f, newArgs],
        tensorVar = args[[tensorPos, 1]],
        newLHS = HoldApply[f, newArgPatts]},
        newDef = ReleaseInnerHold@Hold[
            newLHS := Module[
                {type = TensorType@tensorVar, noMessage = True, result},
                result = Check[newCall, noMessage = False];
                result /; (TensorTypeQ@type && noMessage && newCondition)
            ]
        ]
    ];
    def;
    ReleaseHold@newDef
];

matchAtLevel[expr_, patt_, level_] := expr ~MatchQ~ Nest[{#..}&, patt, level];

ListDepth[a_SparseArray] := ArrayDepth@a;

ListDepth[a_List] := Module[{i = 0, depths},
    While[matchAtLevel[a, _List, i], i++];
    Which[
        matchAtLevel[a, _SparseArray, i] && MatchQ[depths = Flatten@Map[ListDepth, a, {i}], {x_Integer..}],
        i + First@depths,
        matchAtLevel[a, Except[_List | _SparseArray], i],
        i,
        True,
        Indeterminate
    ]
];

ListDepth[_] := 0;

decompDimPattern[dims_] := With[
    {index = Flatten@Position[dims, Verbatim@Blank[], {1}, 1, Heads -> False]},
    HoldPattern[
        expr : {dims..} /;
            And @@ OddQ@expr[[1, index]] && (Length@expr == 1 || Differences@expr[[All, index]] ~MatchQ~ {{2 ..}..})
    ]
];

assumeDecomposition = {"TensorDecompositionMatrix", "TensorDecomposition", "OneByOne"} -> {"TensorDecomposition", "OneByOne", "Operator"};

TensorType[a : _List | _SparseArray, default_List : {}] := TensorType[a, ListDepth@a, default];

TensorType[a : _List | _SparseArray, depth : _Integer | Indeterminate, default_List : {}] :=
    Switch[{depth, Dimensions@a, Dimensions /@ a},
        {1, _, _},
        {"Cartesian"},
        {2, {1, 1}, _},
        {"OneByOne"},
        {2, {1, n_}, _},
        {"Covariant"},
        {2, {n_, 1}, _},
        {"Contravariant"},
        {2, {n_, n_}, _},
        {"Operator"},
        {3, {1, 1, 1}, _},
        {"TensorDecomposition", "OneByOne"}, (* Also could be {"Cartesian", "OneByOne"} *)
        {3, {1, 1, n_}, _} | {3, {n_, 1}, decompDimPattern@{1, _}},
        {"TensorDecomposition", "Covariant"},
        {3, {1, n_, 1}, _} | {3, {n_}, decompDimPattern@{_, 1}},
        {"TensorDecomposition", "Contravariant"},
        {3, {1, n_, n_}, _} | {3, {_}, decompDimPattern@{_, _}},
        {"TensorDecomposition", "Operator"},
        {3, {n_, 1, 1}, _},
        {"Cartesian", "OneByOne"},
        {3, {n_, 1, m_}, _},
        {"Cartesian", "Covariant"},
        {3, {n_, m_, 1}, _},
        {"Cartesian", "Contravariant"},
        {3, {n_, m_, m_}, _},
        {"Cartesian", "Operator"},
        {4, {1, 1, 1, 1}, _},
        {"OneByOne", "OneByOne"},
        {4, {1, 1, 1, n_}, _},
        {"OneByOne", "Covariant"},
        {4, {1, 1, n_, 1}, _},
        {"OneByOne", "Contravariant"},
        {4, {1, 1, n_, n_}, _},
        {"OneByOne", "Operator"},
        {4, {1, n_, 1, 1}, _},
        {"Covariant", "OneByOne"},
        {4, {n_, 1, 1, 1}, _},
        {"Contravariant", "OneByOne"},
        {4, {1, n_, 1, m_}, _},
        {"Covariant", "Covariant"},
        {4, {1, n_, m_, 1}, _},
        {"Covariant", "Contravariant"},
        {4, {1, n_, m_, m_}, _},
        {"Covariant", "Operator"},
        {4, {n_, 1, 1, m_}, _},
        {"Contravariant", "Covariant"},
        {4, {n_, 1, m_, 1}, _},
        {"Contravariant", "Contravariant"},
        {4, {n_, 1, m_, m_}, _},
        {"Contravariant", "Operator"},
        {5, {n_, n_, 1, 1, 1}, _},
        {"TensorDecompositionMatrix", "TensorDecomposition", "OneByOne"}, (* 1x1 case could also be {"TensorDecomposition", "OneByOne", "Operator"}.*)
        {5, {1, 1, 1, n_, n_}, _},
        {"TensorDecomposition", "OneByOne", "Operator"},
        {5, {1, 1, n_, m_, m_}, _} | {5, {n_, 1}, decompDimPattern@{1, _, m_, m_}},
        {"TensorDecomposition", "Covariant", "Operator"},
        {5, {1, n_, 1, m_, m_}, _} | {5, {n_}, decompDimPattern@{_, 1, m_, m_}},
        {"TensorDecomposition", "Contravariant", "Operator"},
        {5 | 7, _, _},
        Switch[
            {Dimensions@a, Flatten[Map[TensorType[#, depth - 2, default]&, a, {2}] /. assumeDecomposition, {{3}, {1, 2}}]},
            (* For now, consider any top level 1x1 that contains a TensorDecomposition to be a TensorDecompositionMatrix *)
            {{n_, n_, ___}, {{"TensorDecomposition"..}, {"OneByOne"..}, {"Operator"..}}},
            {"TensorDecompositionMatrix", "TensorDecomposition", "OneByOne", "Operator"},
            {{n_, n_, ___}, {{"TensorDecomposition"..}, {("Covariant" | "OneByOne")..}}},
            {"TensorDecompositionMatrix", "TensorDecomposition", "Covariant"},
            {{n_, n_, ___}, {{"TensorDecomposition"..}, {("Covariant" | "OneByOne")..}, {"Operator"..}}},
            {"TensorDecompositionMatrix", "TensorDecomposition", "Covariant", "Operator"},
            {{n_, n_, ___}, {{"TensorDecomposition"..}, {("Contravariant" | "OneByOne")..}}},
            {"TensorDecompositionMatrix", "TensorDecomposition", "Contravariant"},
            {{n_, n_, ___}, {{"TensorDecomposition"..}, {("Contravariant" | "OneByOne")..}, {"Operator"..}}},
            {"TensorDecompositionMatrix", "TensorDecomposition", "Contravariant", "Operator"},
            {{1, n_, ___}, {{"TensorDecomposition"..}, {"OneByOne"..}}},
            {"Covariant", "TensorDecomposition", "OneByOne"},
            {{1, n_, ___}, {{"TensorDecomposition"..}, {("Covariant" | "OneByOne")..}}},
            {"Covariant", "TensorDecomposition", "Covariant"},
            {{1, n_, ___}, {{"TensorDecomposition"..}, {("Contravariant" | "OneByOne")..}}},
            {"Covariant", "TensorDecomposition", "Contravariant"},
            {{1, n_, ___}, {{"TensorDecompositionMatrix"..}, {"TensorDecomposition"..}, {"OneByOne"..}}},
            {"Covariant", "TensorDecompositionMatrix", "TensorDecomposition", "OneByOne"},
            {{1, n_, ___}, {{"TensorDecompositionMatrix"..}, {"TensorDecomposition"..}, {("Covariant" | "OneByOne")..}}},
            {"Covariant", "TensorDecompositionMatrix", "TensorDecomposition", "Covariant"},
            {{1, n_, ___}, {{"TensorDecompositionMatrix"..}, {"TensorDecomposition"..}, {("Contravariant" | "OneByOne")..}}},
            {"Covariant", "TensorDecompositionMatrix", "TensorDecomposition", "Contravariant"},
            {{n_, 1, ___}, {{"TensorDecomposition"..}, {"OneByOne"..}}},
            {"Contravariant", "TensorDecomposition", "OneByOne"},
            {{n_, 1, ___}, {{"TensorDecomposition"..}, {("Covariant" | "OneByOne")..}}},
            {"Contravariant", "TensorDecomposition", "Covariant"},
            {{n_, 1, ___}, {{"TensorDecomposition"..}, {("Contravariant" | "OneByOne")..}}},
            {"Contravariant", "TensorDecomposition", "Contravariant"},
            {{n_, 1, ___}, {{"TensorDecompositionMatrix"..}, {"TensorDecomposition"..}, {"OneByOne"..}}},
            {"Contravariant", "TensorDecompositionMatrix", "TensorDecomposition", "OneByOne"},
            {{n_, 1, ___}, {{"TensorDecompositionMatrix"..}, {"TensorDecomposition"..}, {("Covariant" | "OneByOne")..}}},
            {"Contravariant", "TensorDecompositionMatrix", "TensorDecomposition", "Covariant"},
            {{n_, 1, ___}, {{"TensorDecompositionMatrix"..}, {"TensorDecomposition"..}, {("Contravariant" | "OneByOne")..}}},
            {"Contravariant", "TensorDecompositionMatrix", "TensorDecomposition", "Contravariant"},
            _,
            If[default === {}, Message[TensorType::unrec, a]];
            default
        ],
        _,
        If[default === {}, Message[TensorType::unrec, a]];
        default
    ];

TensorType[_, default_List : {}] := {"Scalar"};

CovariantQ[expr_] := MatchQ[TensorType[expr, {""}], {"Covariant", ___}];

ContravariantQ[expr_] := MatchQ[TensorType[expr, {""}], {"Contravariant", ___}];

SphericalTensorQ[expr_] := MatchQ[TensorType[expr, {""}], {"Covariant" | "Contravariant", ___}];

CartesianQ[expr_] := MatchQ[TensorType[expr, {""}], {"Cartesian", ___}];

OperatorQ[expr_] := MatchQ[TensorType[expr, {""}], {"Operator" | "OneByOne"}];

CovariantDecompositionQ[expr_] := MatchQ[TensorType[expr, {""}], {"TensorDecomposition", "Covariant", ___}];

ContravariantDecompositionQ[expr_] := MatchQ[TensorType[expr, {""}], {"TensorDecomposition", "Contravariant", ___}];

TensorDecompositionQ[expr_] := MatchQ[TensorType[expr, {""}], {"TensorDecomposition", ___}];

TensorDecompositionMatrixQ[expr_] := MatchQ[TensorType[expr, {""}], {"TensorDecompositionMatrix", ___}];

CovariantDecompositionMatrixQ[expr_] := MatchQ[TensorType[expr, {""}], {"TensorDecompositionMatrix", "TensorDecomposition", "Covariant", ___}];

ContravariantDecompositionMatrixQ[expr_] := MatchQ[TensorType[expr, {""}], {"TensorDecompositionMatrix", "TensorDecomposition", "Contravariant", ___}];

OneByOneQ[expr_] := MatchQ[TensorType[expr, {""}], {"OneByOne", ___}];

ScalarQ[expr_] := MatchQ[TensorType[expr, {""}], {"Scalar"}];

zeroIndexTensors = {"Scalar"};
oneIndexTensors = {"Cartesian", "TensorDecomposition"};
twoIndexTensors = {"Covariant", "Contravariant", "OneByOne"};
operatorTypeTensors = {"Operator", "TensorDecompositionMatrix"};

tensorLevels = Dispatch@Flatten[
    Thread /@ {zeroIndexTensors -> 0, oneIndexTensors -> 1, twoIndexTensors -> 2, operatorTypeTensors -> 2}
];

(TensorDepth[{a_, type : TensorTypePattern[]}, n_ : -1] := Total[
    If[NonNegative@n, Take[type, n], Drop[type, n + 1]] /. tensorLevels
]) // AddTensorTypeToCall;

(ComponentMap[f_, {a_, type : TensorTypePattern[]}, n_ : 1] := Map[f, a, {TensorDepth[{a, type}, n]}]) // AddTensorTypeToCall;

(* Tensor rank *)

Rank[{a_, type : {"Scalar"} | {"OneByOne", ___}}] := 0;

Rank[{a_, {"Covariant", ___}}] := (Length@First@a - 1) / 2;

Rank[{a_, {"Contravariant" | "Cartesian" | "Operator", ___}}] := (Length@a - 1) / 2;

Rank[{a_, type : {"TensorDecomposition", __} | {"TensorDecompositionMatrix", "TensorDecomposition", __}}] :=
    ComponentMap[Rank[{#, Rest@type}]&, {a, type}, 1];

(* Display *)

TensorForm[a_] := Module[
    {type = TensorType[a, {"Scalar"}]},
    Switch[{TensorDepth[{a, type}], type},
        {0 | 1, _},
        a,
        {2 | 4, _},
        MatrixForm@a,
        {3, _} | {5 | 7, {"TensorDecomposition", __}},
        MatrixForm /@ a,
        {5 | 7, {_, "TensorDecomposition", __}},
        MatrixForm[Map[MatrixForm, a, {3}], TableDepth -> 2],
        {5 | 7, {_, "TensorDecompositionMatrix", __}},
        MatrixForm[Map[MatrixForm, a, {5}], TableDepth -> 4]
    ] /; TensorTypeQ@type
];

(* Tensor components *)

Components[{a_, type : {Alternatives @@ Join[zeroIndexTensors, oneIndexTensors], ___}}] := a;

Components[{a_, type : {Alternatives @@ Join[twoIndexTensors, operatorTypeTensors], ___}}] :=
    Flatten[a, 1];

Component[{a_, type : {"Scalar"}}] := a;

Component[{a_, type : {"OneByOne", ___} | {"TensorDecompositionMatrix", __}}] /; MatchQ[Dimensions@a, {1, 1, ___}] :=
    a[[1, 1]];

Component[{a_, type : {"Operator"} | {"OneByOne", ___}}, {m1_?HalfIntegerQ, m2_?HalfIntegerQ}] := Module[
    {rank = Rank@{a, type}},
    a[[-(m1 + rank + 1), -(m2 + rank + 1)]]
];

Component[{a_, type : {"Scalar" | "OneByOne", ___}}, 0] :=
    Component@{a, type};

Component[{a_, type : {"Cartesian", ___}}, i_Integer] :=
    a[[i]];

Component[{a_, type : {"TensorDecomposition", __}}, k_Integer] :=
    First@Select[a, Rank[{#, Rest@type}] == k&, 1];

Component[{a_, type : {"Covariant", ___}}, q_?HalfIntegerQ] :=
    a[[1, -(q + Rank@{a, type} + 1)]];

Component[{a_, type : {"Contravariant", ___}}, q_?HalfIntegerQ] :=
    a[[-(q + Rank@{a, type} + 1), 1]];

Component[{a_, type : TensorTypePattern[]}, q_] :=
    Message[Component::wrongspec, type, q];

Component[{a_, type : TensorTypePattern[]}, indices__] /; Length@{indices} <= Length@type :=
    First@Fold[{Component@##, Rest@Last@#1}&, {a, type}, {indices}];

Component[{a_, type : TensorTypePattern[]}, indices__] :=
    Message[Component::extraindices, type, Length@{type}, Length@{indices}];

Component[a : Except[TypedTensorPattern[] | AtomicSystemPattern[]], args__] := Module[
    {type = TensorType@a, noMessage = True, result},
    result = Check[
        QuietCheck[
            Component[{a, type}, args],
            If[Length@{args} == 1,
                Message[Component::noexist, type, args],
                Message[Component::noexistn, type, {args}]
            ],
            {Part::partw}
        ],
        noMessage = False
    ];
    result /; noMessage
];

Component[sys_?AtomicSystemQ, {a_, type : {"Operator"} | {"OneByOne", ___}}, {s1_AtomicState, s2_AtomicState}] :=
    MatrixElement[sys, a, {s1, s2}];

Component[sys_?AtomicSystemQ, {a_, type : {"TensorDecompositionMatrix", __}}, {s1_AtomicState, s2_AtomicState}] :=
    MatrixElement[RemoveZeemanSublevels@sys, a, DeleteParameter[#, M]& /@ {s1, s2}];

(*Component[sys_?AtomicSystemQ, {a_, type : {"Operator" | "OneByOne"}}, crit1_ -> crit2_] :=*)
    (*MatrixElement[sys, a, crit1 -> crit2];*)

(*Component[sys_?AtomicSystemQ, {a_, type : {"TensorDecompositionMatrix", __}}, crit1_ -> crit2_] :=*)
    (*MatrixElement[RemoveZeemanSublevels@sys, a, crit1 -> crit2];*)

Component[sys_?AtomicSystemQ, {a_, type : TensorTypePattern[]}, k_] :=
    Message[Component::wrongspec, type, k];

Component[sys_?AtomicSystemQ, {a_, type : TensorTypePattern[]}, indices__] /; Length@{indices} <= Length@type :=
    First@Fold[
        {
            If[MatchQ[#1[[2, 1]], "Operator" | "TensorDecompositionMatrix"],
                Component[sys, ##],
                Component[##]
            ],
            Rest@Last@#1
        }&,
        {a, type},
        {indices}
    ];

Component[sys_?AtomicSystemQ, {a_, type : TensorTypePattern[]}, indices__] :=
    Message[Component::extraindices, type, Length@{type}, Length@{indices}];

Component[sys_?AtomicSystemQ, a : Except@TypedTensorPattern[], args__] := Module[
    {type = TensorType@a, noMessage = True, result},
    result = Check[
        QuietCheck[
            Component[sys, {a, type}, args],
            If[Length@{args} == 1,
                Message[Component::noexist, type, args],
                Message[Component::noexistn, type, {args}]
            ],
            {Part::partw, Extract::psl1}
        ],
        noMessage = False
    ];
    result /; noMessage
];

LowerTriangle[{a_, type : {"Operator"}}] := Flatten@MapIndexed[Take[#1, First@#2]&, a];

LowerTriangle[{a_, type : {"TensorDecompositionMatrix", __}}] := Module[
    {rank = Rank@a},
    Flatten@Table[
        Component[{a[[i, j]], Rest@type}, k, q],
        {i, Length@a},
        {j, i},
        {k, rank[[i, j]]},
        {q, -k, If[i == j, 0, k]}
    ]
];

UpperTriangle[{a_, type : {"Operator"}}] := LowerTriangle@Transpose@a;

UpperTriangle[{a_, type : {"TensorDecompositionMatrix", __}}] := Module[
    {rank = Rank@a},
    Flatten@Table[
        Component[{a[[i, j]], Rest@type}, k, q],
        {i, Length@a},
        {j, i, Length@a},
        {k, rank[[i, j]]},
        {q, If[i == j, 0, -k], k}
    ]
];

(* Inner and outer products*)

Commutator[a_?OperatorQ, b_?OperatorQ] := a.b - b.a;

Anticommutator[a_?OperatorQ, b_?OperatorQ] := (a.b + b.a);

DirectProduct[f_, a_List, b_List] := Outer[f, a, b, 1];

DirectProduct[a_List, b_List] := DirectProduct[Times, a, b];

DirectSelfProduct[f_, a_List] := DirectProduct[f, a, a];

DirectSelfProduct[a_List] := DirectProduct[a, a];

ScalarizeOneByOnes[a_] := Module[{type = TensorType@a, pos},
    pos = Flatten@Position[type, "OneByOne"];
    Fold[ComponentMap[#[[1, 1]]&, {#1, type}, #2]&, a, Reverse@pos - 1] /; TensorTypeQ@type
];

SetAttributes[QuietCheck, HoldAll];
QuietCheck[expr_, failexpr_, msgs_] := Quiet[Check[expr, failexpr, msgs], msgs];

Options[MatrixMultiply] = {ScalarizeOneByOnes -> True, SumRanks -> True, "MultiplyLevel" -> Automatic, "AllowCartesianTimesOperator" -> False};

OptionProccessor[ScalarizeOneByOnes] = OptionProccessor[SumRanks] = TrueQ;

MatrixMultiply[a_, b_, c : Except[_?OptionQ].., opts : OptionsPattern[]] :=
    MatrixMultiply[MatrixMultiply[a, b, opts], c, opts];

(MatrixMultiply[a_, b_, OptionsPattern[]] := Module[
    {scalarizeOneByOnes, sumRanks, aType = TensorType@a, bType = TensorType@b, noMessage = True, result},
    result = Check[
        QuietCheck[
            matrixMultiply[{a, aType}, {b, bType}, OptionValue@"MultiplyLevel", sumRanks, OptionValue@"AllowCartesianTimesOperator"] //
                If[scalarizeOneByOnes, ScalarizeOneByOnes, Identity],
            Message[MatrixMultiply::dotsh, aType, bType],
            {Dot::dotsh, MapThread::mptc, TensorType::unrec}
        ],
        noMessage = False
    ];
    result /; noMessage
]) // ProcessOptions;

(*Forbidden combinations*)

matrixMultiply[
    {a_, aType : TensorTypePattern[]},
    {b_, bType : TensorTypePattern[]},
    __
] /; MatchQ[{aType, bType},
    {
        {"Cartesian"},
        {"Covariant" | "Contravariant", ___}
    } | {
        {"Covariant" | "Contravariant", ___},
        {"Cartesian"}
    } | {
        {"Covariant"} | {"Cartesian", "Operator" | "Covariant"},
        {"Covariant"} | {"Covariant" | "Cartesian", "Covariant"}
    } | {
        {"Contravariant"} | {"Contravariant" | "Cartesian", "Contravariant"},
        {"Contravariant"} | {"Cartesian", "Operator" | "Contravariant"}
    } | {
        {"Operator"},
        {"Covariant"} | {"Covariant", "Covariant"}
    } | {
        {"Contravariant"} | {"Contravariant", "Contravariant"},
        {"Operator"}
    } | {
        {"Cartesian", _},
        {"Covariant" | "Contravariant", _}
    } | {
        {"Covariant" | "Contravariant", _},
        {"Cartesian", _}
    } | {
        {"Covariant", _},
        {"Covariant", _}
    } | {
        {"Contravariant", _},
        {"Covariant" | "Contravariant", _}
    }
] := Message[MatrixMultiply::nomult, aType, bType];

(*Product involving a scalar*)

matrixMultiply[{a_, {"Scalar"}}, {b_, bType : TensorTypePattern[]}, __] := a b;
matrixMultiply[{a_, bType : TensorTypePattern[]}, {b_, {"Scalar"}}, __] := a b;

(*Product involving a Cartesian vector*)

matrixMultiply[{a_, {"Cartesian"}}, {b_, {"Cartesian"}}, __] := a.b;

matrixMultiply[
    {a_, {"Cartesian" | "Operator"}},
    {b_, {"Cartesian" | "Operator"}},
    _,
    _,
    True
] := a.b;

matrixMultiply[
    {a_, {"Cartesian"}},
    {b_, {"Cartesian", "Operator" | "Covariant" | "Contravariant" | "OneByOne"}},
    __
] := a.b;

matrixMultiply[
    {a_, {"Cartesian", "Operator" | "Covariant" | "Contravariant" | "OneByOne"}},
    {b_, {"Cartesian"}},
    __
] := Total@MapThread[Times, {a, b}];

(*Product involving a 1x1, operator, covariant vector (or bra), or contravariant vector (or ket)*)

matrixMultiply[
    {a_, aType : {"Operator" | "Covariant" | "Contravariant" | "OneByOne"}},
    {b_, bType : {"Operator" | "Covariant" | "Contravariant" | "OneByOne"}},
    __
] := a.b;

(*For MultiplyLevel -> Automatic, assume we're multiplying by co/contravariant classical vector rather than bra vector*)
matrixMultiply[
    {a_, {"Covariant" | "OneByOne"}},
    {b_, {"Contravariant" | "OneByOne", "Operator" | "Covariant" | "Contravariant" | "OneByOne"}},
    Automatic | Top,
    __
] := a.b;

matrixMultiply[
    {a_, aType : {"Covariant" | "OneByOne", "Operator" | "Covariant" | "Contravariant" | "OneByOne"}},
    {b_, bType : {"Contravariant" | "OneByOne"}},
    Automatic | Top,
    __
] := Total@MapThread[Times, {Components@{a, aType}, Components@{b, bType}}];

matrixMultiply[
    {a_, {"Operator" | "Covariant" | "OneByOne"}},
    {b_, bType : {"Covariant" | "Contravariant", "Operator" | "Contravariant" | "OneByOne"}},
    Bottom,
    __
] := ComponentMap[a.#&, {b, bType}];

matrixMultiply[
    {a_, {"Covariant" | "Contravariant" | "OneByOne", "Operator" | "Covariant" | "OneByOne"}},
    {b_, {"Operator" | "Contravariant" | "OneByOne"}},
    Bottom,
    __
] := a.b;

matrixMultiply[
    {a_, {"Operator" | "Covariant" | "OneByOne"}},
    {b_, {"Contravariant" | "OneByOne", "Operator" | "Covariant" | "Contravariant" | "OneByOne"}},
    Automatic | Top,
    __
] := a.b;

matrixMultiply[
    {a_, {"Covariant" | "OneByOne", "Operator" | "Covariant" | "Contravariant" | "OneByOne"}},
    {b_, {"Operator" | "Contravariant" | "OneByOne"}},
    Automatic | Top,
    __
] := Transpose[Transpose[a, {3, 4, 1, 2}].b, {3, 4, 1, 2}];

matrixMultiply[
    {a_, {"Operator" | "Covariant" | "OneByOne"}},
    {b_, bType : {"Cartesian", "Operator" | "Contravariant" | "OneByOne"}},
    Automatic | Bottom,
    __
] := ComponentMap[a.#&, {b, bType}];

matrixMultiply[
    {a_, {"Cartesian", "Operator" | "Covariant" | "OneByOne"}},
    {b_, {"Operator" | "Contravariant" | "OneByOne"}},
    Automatic | Bottom,
    __
] := a.b;

matrixMultiply[
    {a_, {"Operator"}},
    {b_, {"Cartesian", "Operator" | "Covariant" | "Contravariant" | "OneByOne"}},
    Top,
    _,
    True
] := a.b;

matrixMultiply[
    {a_, {"Cartesian", "Operator" | "Covariant" | "Contravariant" | "OneByOne"}},
    {b_, {"Operator"}},
    Top,
    _,
    True
] := Transpose[Transpose[a, {3, 1, 2}].b, {2, 3, 1}];

matrixMultiply[
    {a_, aType : {"Contravariant" | "OneByOne"}},
    {b_, bType : {"Cartesian" | "Covariant" | "Contravariant" | "OneByOne", "Covariant" | "OneByOne"}},
    Automatic | Bottom,
    __
] := ComponentMap[a.#&, {b, bType}];

matrixMultiply[
    {a_, {"Cartesian" | "Covariant" | "Contravariant" | "OneByOne", "Contravariant" | "OneByOne"}},
    {b_, {"Covariant" | "OneByOne"}},
    Automatic | Bottom,
    __
] := a.b;

(*Product involving a Cartesian vector of operators, bras, or kets*)

matrixMultiply[
    {a_, {"Cartesian", "Operator" | "Covariant" | "Contravariant" | "OneByOne"}},
    {b_, {"Cartesian", "Operator" | "Covariant" | "Contravariant" | "OneByOne"}},
    __
] := Total@MapThread[Dot, {a, b}];

(*Product involving a co/contravariant vector of operators, bras, or kets*)

matrixMultiply[
    {a_, aType : {"Covariant" | "Contravariant" | "OneByOne", "Operator" | "Covariant" | "Contravariant" | "OneByOne"}},
    {b_, bType : {"Covariant" | "Contravariant" | "OneByOne", "Operator" | "Covariant" | "Contravariant" | "OneByOne"}},
    __
] := Total@MapThread[Dot, {Components@{a, aType}, Components@{b, bType}}];

(*Product involving tensor decompositions*)

matrixMultiply[
    {a_, aType : {"TensorDecomposition", "Covariant" | "Operator" | "OneByOne", ___}},
    {b_, bType : {"TensorDecomposition", "Contravariant" | "Operator" | "OneByOne", ___}},
    multiplyLevel_,
    sumRanks_,
    _
] := If[sumRanks && MatchQ[{aType, bType}[[All, 2]], {"Covariant" | "OneByOne", "Contravariant" | "OneByOne"}], Total, Identity]@
    MapThread[matrixMultiply[{#1, Rest@aType}, {#2, Rest@bType}, multiplyLevel, sumRanks, False]&, {a, b}];

matrixMultiply[
    {a_, aType : {"TensorDecompositionMatrix", "TensorDecomposition", "Covariant" | "Operator" | "OneByOne", ___}},
    {b_, bType : {"TensorDecompositionMatrix", "TensorDecomposition", "Contravariant" | "Operator" | "OneByOne", ___}},
    multiplyLevel_,
    sumRanks_,
    _
] := If[sumRanks && MatchQ[{aType, bType}[[All, 3]], {"Covariant" | "OneByOne", "Contravariant" | "OneByOne"}], Total[#, 2]&, Identity]@
    MapThread[matrixMultiply[{#1, Rest@aType}, {#2, Rest@bType}, multiplyLevel, sumRanks, False]&, {a, b}, 2];

(*Any cases we haven't addressed*)

matrixMultiply[
    {a_, aType : TensorTypePattern[]},
    {b_, bType : TensorTypePattern[]},
    __
] := Message[MatrixMultiply::nomult, aType, bType];

CenterDot = MatrixMultiply;

Wedge = MatrixMultiply[##, "MultiplyLevel" -> Top]&;

Vee = MatrixMultiply[##, "MultiplyLevel" -> Bottom]&;

(* Conjugation *)

ExplicitConjugate[a_] := a /. Complex[x_, y_] -> Complex[x, -y];

EC = ExplicitConjugate;

SuperStar = ExplicitConjugate;

conjugateTranspose[conjugate_, {a_, type : {"Scalar" | "Cartesian" | "OneByOne"}}] := conjugate@a;

conjugateTranspose[conjugate_, {a_, type : {"Covariant" | "Contravariant" | "Operator"}}] := Transpose@conjugate@a;

conjugateTranspose[conjugate_, {a_, type : {"Cartesian" | "TensorDecomposition" | "OneByOne" | "TensorDecompositionMatrix", __}}] :=
    ComponentMap[conjugateTranspose[conjugate, {#, Rest@type}]&, {a, type}];

conjugateTranspose[conjugate_, {a_, type : {"Covariant" | "Contravariant" | "Operator", __}}] :=
    Transpose@ComponentMap[conjugateTranspose[conjugate, {#, Rest@type}]&, {a, type}];

(ExplicitConjugateTranspose[{a_, type : TensorTypePattern[]}] := conjugateTranspose[ExplicitConjugate, {a, type}]) // AddTensorTypeToCall;

ECT = ExplicitConjugateTranspose;

(GeneralizedConjugateTranspose[{a_, type : TensorTypePattern[]}] := conjugateTranspose[Conjugate, {a, type}]) // AddTensorTypeToCall;

GCT = GeneralizedConjugateTranspose;

SuperDagger = GeneralizedConjugateTranspose;

(*Dual*)

(*Switch representation from covariant to contravariant, or vice versa. [Var88, Sec. 3.1.2, Eq. (10)]*)
(*	Subscript[A, q]=(-1)^q A^-q,	A^q=(-1)^q Subscript[A, -q]*)
(*where A is any real irreducible tensor, including real vectors.*)

DualMatrix[rank_?HalfIntegerQ] := Reverse@DiagonalMatrix@Reverse@Table[(-1)^q, {q, -rank, rank}]; (*Check if half-integer rank makes sense.*)

Dual[{a_, type : {"OneByOne"} | {"Cartesian", ___}}] := a;

Dual[{a_, type : {"TensorDecompositionMatrix" | "TensorDecomposition", __}}] := ComponentMap[Dual, {a, type}];

Dual[{a_, type : {"Contravariant", ___}}] := Transpose[DualMatrix@Rank@a\[Wedge]a];

Dual[{a_, type : {"Covariant", ___}}] := Transpose[a\[Wedge]DualMatrix@Rank@a];

ToCovariant[{a_, type : {"Covariant", ___} | {"OneByOne"}}] := a;

ToCovariant[{a_, type : {"Contravariant", ___}}] := Dual@{a, type};

ToCovariant[{a_, type : {"TensorDecompositionMatrix" | "TensorDecomposition", __}}] := ComponentMap[ToCovariant, {a, type}];

ToContravariant[{a_, type : {"Contravariant", ___} | {"OneByOne"}}] := a;

ToContravariant[{a_, type : {"Covariant", ___}}] := Dual@{a, type};

ToContravariant[{a_, type : {"TensorDecompositionMatrix" | "TensorDecomposition", __}}] := ComponentMap[ToContravariant, {a, type}];

(*Complex conjugate of an irreducible tensor [Var88, Sec. 1.2.1, Eq. (8)]*)
(*	Subscript[( *)
(*\!\(\*SuperscriptBox["A", "*"]\)), q]=*)
(*\!\(\*SuperscriptBox[*)
(*RowBox[{"(", *)
(*SuperscriptBox["A", "q"], ")"}], "*"]\)=(-1)^q *)
(*\!\(\*SuperscriptBox[*)
(*RowBox[{"(", *)
(*SubscriptBox["A", *)
(*RowBox[{"-", "q"}]], ")"}], "*"]\)*)
(*For a vector, this is equivalent to *)
(*	ToSpherical[Conjugate[ToCartesian[A]]]*)

TensorConjugate[a_?SphericalTensorQ] := Transpose@Conjugate@Dual@a;

TensorConjugate[a_?CartesianQ] := Conjugate@a;

TensorNorm[a_ /; SphericalTensorQ[a] || CartesianQ[a]] :=
    Sqrt[ToCovariant@a\[CenterDot]ToContravariant@a];

TensorNorm[a_?TensorDecompositionQ] := TensorNorm /@ a;

(*Spherical to Cartesian conversion*)

(*Any vector A can be expanded in terms of basis vectors:*)
(*	A=\[Sum]A^\[Mu] Subscript[e, \[Mu]]=\[Sum]Subscript[A, \[Mu]] e^\[Mu]*)
(*Subscript[A, \[Mu]] are the covariant components*)
(*	Subscript[A, \[Mu]]=A\[CenterDot]Subscript[e, \[Mu]],*)
(*and A^\[Mu] are the contravariant components*)
(*	A^\[Mu]=A\[CenterDot]e^\[Mu].*)

(*Convert the spherical components of a vector [Var88, Sec. 1.2.1, Eq. (11)] or a set of spherical basis vectors [Var88, Sec. 1.1.5, Eq. (30)] to cartesian representation. The relations are the same in either case.*)
(* The behavior seems pretty arbitrary for kappa > 1. Maybe should set it up so that the components of the tensor in the "Cartesian" basis are all real *)

ToCartesianMatrix[rank_?HalfIntegerQ] :=
    Flatten[{
        Table[
            {Table[Switch[j, i, -1, -i, 1, _, 0], {j, rank, -rank, -1}],
                Table[Switch[j, i, -I, -i, -I, _, 0], {j, rank, -rank, -1}]},
            {i, rank, 1/2, -1}] / Sqrt[2],
        {If[IntegerQ@rank, {Table[KroneckerDelta[j], {j, rank, -rank, -1}]}, {}]}
    }, 2];

ToCartesian[{a_, type : {"OneByOne", ___}}] := First@a;

ToCartesian[{a_, {"Cartesian", ___} | {"Scalar"}}] := a;

ToCartesian[{a_, {"Contravariant", ___}}] := Components[ToCartesianMatrix@Rank@a\[Wedge]a];

ToCartesian[{a_, {"Covariant", ___}}] := Components[a\[Wedge]ECT@ToCartesianMatrix@Rank@a];

ToContravariant[{a_, type : {"Cartesian", ___}}] :=
    Transpose@{MatrixMultiply[ECT@ToCartesianMatrix@Rank@a, a, "MultiplyLevel" -> Top, "AllowCartesianTimesOperator" -> True]};

ToCovariant[{a_, type : {"Cartesian", ___}}] :=
    {MatrixMultiply[a, ToCartesianMatrix@Rank@a, "MultiplyLevel" -> Top, "AllowCartesianTimesOperator" -> True]};

ToCartesian[sys_?AtomicSystemQ, op_?OperatorQ] := Module[
    {mat = BlockDiagonal[ToCartesianMatrix /@ F /@ RemoveZeemanSublevels@sys]},
    mat\[CenterDot]op\[CenterDot]ECT[mat]
];

(*Tensor product*)

(*Irreducible tensor product Subscript[\!\(TraditionalForm\`{\**)
(*SubscriptBox[*)
(*StyleBox["A",*)
(*FontWeight->"Bold",*)
(*FontSlant->"Plain"], *)
(*SubscriptBox["\[Kappa]", "1"]]\[CircleTimes]\**)
(*SubscriptBox[*)
(*StyleBox["B",*)
(*FontWeight->"Bold",*)
(*FontSlant->"Plain"], *)
(*SubscriptBox["\[Kappa]", "2"]]}\), \[Kappa]] [Var88, Sec. 3.1.7, Eqs. (20,21)]*)
(*The tensor product of rank 0 is related to the scalar product by [Var88, Sec. 3.1.8, Eq. (35)]*)
(*	Subscript[A, \[Kappa]]\[CenterDot]Subscript[B, \[Kappa]]=(-1)^\[Kappa] Sqrt[2\[Kappa]+1] Subscript[{Subscript[A, \[Kappa]]\[CircleTimes]Subscript[B, \[Kappa]]}, 00]*)

SphericalTensorProduct[a_?SphericalTensorQ, b_?SphericalTensorQ, {k_}] :=
    {Reverse@Table[SphericalTensorProduct[a, b, {k, q}], {q, -k, k}]};

SphericalTensorProduct[a_?CovariantQ, b_?CovariantQ, {k_, q_}] := Module[{aRank = Rank@a, bRank = Rank@b},
    Sum[
        Sum[
            CG[{aRank, q1}, {bRank, q2}, {k, q}] Component[a, q1] Component[b, q2],
            {q2, -bRank, bRank}
        ],
        {q1, -aRank, aRank}
    ]
];

(*Wrappers and catch-alls*)

AddTensorTypeToCall[#[{a_, type : TensorTypePattern[]}] := Null /; Message[#::wrongtype, type]]& /@
    {Component, Components, ToCartesian, ToCovariant, ToContravariant, Dual, LowerTriangle, UpperTriangle, Rank,
        conjugateTranspose};

End[];

EndPackage[];
